# -*- coding: utf-8 -*-
# This file is part of fedimg.
# Copyright (C) 2014-2017 Red Hat, Inc.
#
# fedimg is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# fedimg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with fedimg; if not, see http://www.gnu.org/licenses,
# or write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
# Authors:  David Gay <dgay@redhat.com>
#           Sayan Chowdhury <sayanchowdhury@fedoraproject.org>
"""
This module checks for the active services (configured through the fedimg
configuration file) and call the main method for the service.
"""

import logging

from fedimg.config import ACTIVE_SERVICES
from fedimg.services.ec2.ec2initiate import main as ec2main
from fedimg.services.ec2.ec2copy import main as ec2copy
from fedimg.config import AWS_ACCESS_ID, AWS_SECRET_KEY
from fedimg.config import AWS_BASE_REGION, AWS_REGIONS


LOG = logging.getLogger(__name__)


def upload(pool, urls, *args, **kwargs):
    """
    Takes a list (urls) of one or more .raw.xz image files and
    sends them off to cloud services for registration. The upload
    jobs threadpool must be passed as `pool`.

    Args:
        pool (multithreading.pool.ThreadPool): The thread pool to parallelize
        the uploads.
        urls (list): List of cloud image urls.
    """

    active_services = ACTIVE_SERVICES
    compose_id = kwargs.get('compose_id')
    push_notifications = kwargs.get('push_notifications')

    if 'aws' in active_services:
        LOG.info('Starting to process AWS EC2Service.')
        images_metadata = ec2main(
            urls,
            AWS_ACCESS_ID,
            AWS_SECRET_KEY,
            [AWS_BASE_REGION],
            compose_id=compose_id,
            push_notifications=push_notifications,
        )
        for image_metadata in images_metadata:
            image_id = image_metadata['image_id']
            aws_regions = list(set(AWS_REGIONS) - set([AWS_BASE_REGION]))
            ec2copy(
                aws_regions,
                AWS_ACCESS_ID,
                AWS_SECRET_KEY,
                image_ids=[image_id],
                push_notifications=push_notifications,
                compose_id=compose_id
            )
        LOG.info('AWS EC2Service process is completed.')



# -*- coding: utf-8 -*-
# This file is part of fedimg.
# Copyright (C) 2014-2017 Red Hat, Inc.
#
# fedimg is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# fedimg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with fedimg; if not, see http://www.gnu.org/licenses,
# or write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
# Authors:  Sayan Chowdhury <sayanchowdhury@fedoraproject.org>
#
"""
I don't know what to write here for documentation
"""

import logging
import os
import shutil

from itertools import product as itertools_product

import fedimg.messenger

from fedimg.config import AWS_VOLUME_TYPES, AWS_BASE_REGION
from fedimg.config import AWS_ROOT_VOLUME_SIZE
from fedimg.services.ec2.ec2imguploader import EC2ImageUploader
from fedimg.services.ec2.ec2imgpublisher import EC2ImagePublisher
from fedimg.utils import get_virt_types_from_url, get_source_from_image
from fedimg.utils import get_image_name_from_image, get_file_arch

LOG = logging.getLogger(__name__)


def main(image_urls, access_id, secret_key, regions, volume_types=None,
         volume_via_s3=True, ex_virt_types=None, push_notifications=False,
         compose_id=None):
    """
    The `ec2.ec2initiate.main` function iterates over the image urls and start
    uploading the image to the specified regions. The `image_urls`,
    `access_id`, and `regions` are the required params. `volume_types`,
    `ex_virt_types` are optional arguments, if not passed the values are picked
    up from the fedimg configuration.
    `volume_via_s3`, `push_notifications`, and `compose_id` are optional params
    with default values.

    Args:
        image_urls (list): List of the image urls to create AMIs. (reques
        access_id (str): AWS EC2 access id
        secret_key (str): AWS_EC2 secret key
        regions (list): List of AWS regions the AMI to be uploaded.
        volume_types (list): List of supported volumes for the AMIs to
            be created.
        volume_via_s3 (bool): If `True`, the images are uploaded via s3 method
            else using creating builder instances.
        ex_virt_types (list): List of the supported virts for the AMIs to
            be created.
        push_notifications (bool): If `True` the messages will be pushed to
            fedmsg, else skipped.
        compose_id: id of the current compose in process.
    """

    root_volume_size = AWS_ROOT_VOLUME_SIZE
    published_images = []

    if volume_types is None:
        volume_types = AWS_VOLUME_TYPES

    if regions is None:
        regions = [AWS_BASE_REGION]

    for image_url in image_urls:

        # If the virt types is not provided then select the supported virt
        # types from the image.
        if ex_virt_types is None:
            virt_types = get_virt_types_from_url(image_url)
        else:
            virt_types = ex_virt_types

        try:
            source = get_source_from_image(image_url)
            if not source:
                raise ValueError

            image_architecture = get_file_arch(image_url)

            uploader = EC2ImageUploader(
                compose_id=compose_id,
                access_key=access_id,
                secret_key=secret_key,
                root_volume_size=root_volume_size,
                image_architecture=image_architecture,
                volume_via_s3=volume_via_s3,
                push_notifications=push_notifications,
                image_url=image_url
            )

            publisher = EC2ImagePublisher(
                compose_id=compose_id,
                access_key=access_id,
                secret_key=secret_key,
                push_notifications=push_notifications,
            )

            combinations = itertools_product(
                    *[regions, virt_types, volume_types])
            for region, virt_type, volume_type in combinations:
                uploader.set_region(region)
                LOG.debug('(uploader) Region is set to: %r' % region)

                uploader.set_image_virt_type(virt_type)
                LOG.debug('(uploader) Virtualization type '
                          'is set to: %r' % virt_type)

                image_name = get_image_name_from_image(
                    image_url=image_url,
                    virt_type=virt_type,
                    region=region,
                    volume_type=volume_type
                )
                uploader.set_image_name(image_name)

                uploader.set_image_volume_type(volume_type)
                LOG.debug('(uploader) Volume type is set to: %r' % volume_type)

                uploader.set_availability_zone_for_region()

                if push_notifications:
                    fedimg.messenger.notify(
                        topic='image.upload',
                        msg=dict(
                            image_url=image_url,
                            image_name=image_name,
                            destination=region,
                            service='EC2',
                            status='started',
                            compose=compose_id,
                            extra=dict(
                                virt_type=virt_type,
                                vol_type=volume_type
                            )
                        )
                    )
                image = uploader.create_image(source)

                published_images.extend(publisher.publish_images(
                    region_image_mapping=[(region, image.id)]
                ))
        except Exception as e:
            LOG.debug(e.message)
            #TODO: Implement the clean up of the images if failed.
            # uploader.clean_up(image_id=image.id, delete_snapshot=True)

    shutil.rmtree(os.path.dirname(source))
    return published_images



# -*- coding: utf-8 -*-
# This file is part of fedimg.
# Copyright (C) 2014-2017 Red Hat, Inc.
#
# fedimg is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# fedimg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with fedimg; if not, see http://www.gnu.org/licenses,
# or write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
# Authors:  Sayan Chowdhury <sayanchowdhury@fedoraproject.org>
#
"""
I don't know what to write here for documentation.
"""
import logging

from fedimg.config import AWS_BASE_REGION
from fedimg.services.ec2.ec2imgpublisher import EC2ImagePublisher

LOG = logging.getLogger(__name__)


def main(regions, access_id, secret_key, images=None,
         image_ids=None, push_notifications=False, compose_id=None):

    if images is None and image_ids is None:
        LOG.debug('Either images or image_ids is required')
        raise

    publisher = EC2ImagePublisher(
        compose_id=compose_id,
        access_key=access_id,
        secret_key=secret_key,
        push_notifications=True,
    )

    if image_ids is None and images is not None:
        image_ids = [image.id for image in images]

    for image_id in image_ids:
        copied_images = publisher.copy_images_to_regions(
            image_id=image_id,
            regions=regions,
            base_region=AWS_BASE_REGION)
        copied_images = [elem.values() for elem in copied_images]
        publisher.publish_images(region_image_mapping=copied_images)



# This file is part of fedimg.
# Copyright (C) 2014-2017 Red Hat, Inc.
#
# fedimg is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# fedimg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with fedimg; if not, see http://www.gnu.org/licenses,
# or write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
# Authors:  David Gay <dgay@redhat.com>
#           Ralph Bean <rbean@redhat.com>
#           Sayan Chowdhury <sayanchowdhury@fedoraproject.org>

"""
Utility functions for fedimg.
"""
import logging
log = logging.getLogger("fedmsg")

import functools
import os
import re
import socket
import subprocess
import tempfile

import paramiko
from libcloud.compute.types import Provider
from libcloud.compute.providers import get_driver


def get_file_arch(file_name):
    """ Takes a file name (probably of a .raw.xz image file) and returns
    the suspected architecture of the contained image. If it doesn't look
    like a 32-bit or 64-bit image, None is returned. """
    if file_name.find('x86_64') != -1:
        return 'x86_64'
    else:
        return None


def get_rawxz_urls(location, images):
    """ Iterates through all the images metadata and returns the url of .raw.xz
    files.
    """
    rawxz_list = [f['path'] for f in images if f['path'].endswith('.raw.xz')]
    if not rawxz_list:
        return []

    return map((lambda path: '{}/{}'.format(location, path)), rawxz_list)


def get_virt_types_from_url(url):
    """ Takes a URL to a .raw.xz image file) and returns the suspected
        virtualization type that the image file should be registered as. """
    return ['hvm']


def region_to_driver(region):
    """ Takes a region name (ex. 'eu-west-1') and returns
    the appropriate libcloud provider value. """
    cls = get_driver(Provider.EC2)
    return functools.partial(cls, region=region)


def ssh_connection_works(username, ip, keypath):
    """ Returns True if an SSH connection can me made to `username`@`ip`. """
    ssh = paramiko.SSHClient()
    ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    works = False
    try:
        ssh.connect(ip, username=username,
                    key_filename=keypath)
        works = True
    except (paramiko.BadHostKeyException,
            paramiko.AuthenticationException,
            paramiko.SSHException, socket.error):
        pass
    ssh.close()
    return works


def get_value_from_dict(_dict, *keys):
    for key in keys:
        try:
            _dict = _dict[key]
        except KeyError:
            return None
    return _dict


def external_run_command(command):
    log.debug("Starting the command: %r" % command)
    ret = subprocess.Popen(' '.join(command), stdin=subprocess.PIPE, shell=True,
                           stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                           close_fds=True)
    out, err = ret.communicate()
    log.debug("Finished executing the command: %r" % command)
    retcode = ret.returncode
    return out, err, retcode


def get_item_from_regex(output, regex):
    match = re.search(regex, output)
    if match is None:
        return ''
    else:
        return match.group(1)


def get_file_name_image(image_url):
    return image_url.split('/')[-1]


def get_source_from_image(image_url):
    tmpdir = tempfile.mkdtemp()
    file_name = get_file_name_image(image_url)
    file_path = os.path.join(tmpdir, file_name)

    log.info("[PREP] Preparing temporary directory for download: %r" % tmpdir)
    output, error, retcode = external_run_command([
        'wget',
        image_url,
        '-P',
        tmpdir
    ])

    if retcode != 0:
        return ''

    return file_path


def get_volume_type_from_image(image, region):
    return image.extra['block_device_mapping'][0]['ebs']['volume_type']


def get_virt_type_from_image(image):
    return 'hvm'


def get_image_name_from_image(image_url, virt_type='', region='', respin='0',
                              volume_type=''):

    file_name = get_file_name_image(image_url)
    build_name = file_name.replace('.raw.xz', '')

    return '-'.join(
        [x for x in [build_name, virt_type, region, volume_type, respin] if x])


def get_image_name_from_ami_name(image_name, region):
    name_vt_region, volume_type, respin = image_name.rsplit('-', 2)
    name_vt = name_vt_region.rsplit('-', 3)[:-3][0]

    return '-'.join(
        [x for x in [name_vt, region, volume_type, respin] if x])
